use std::{thread::available_parallelism};

use tokio::{
    net::TcpListener,
    select,
    signal::unix::{signal, SignalKind, Signal},
};

use tokio_native_tls::{native_tls::{Identity, self}, TlsAcceptor};

use crate::{
    handlers::handler::handle_connection,
    handling::{routes::{Route, Uri}, response::{Response, Status}},
};

#[derive(Clone)]
/// Represnts a [MountPoint] that can be mounted in the config
pub struct MountPoint<'a> {
    /// The prefix of the [MountPoint]
    pub mountpoint: Uri<'a>,
    /// All Routes mounted on the [MountPoint]. The Routes are all prefixed by the mountpoints
    /// mountpoint
    pub routes: Vec<Route<'a>>,
}

/// A server configuration that is run
pub struct Config {
    /// Contains an Optional [`Vec<MountPoint>`]. which contains all [MountPoint]s on the Server.
    /// Which contain all the [Route]s
    mountpoints: Option<Vec<MountPoint<'static>>>,
    /// Contains a [tokio::net::TcpListener] that is bound for the server
    address: TcpListener,
    #[cfg(feature = "secure")]
    to_secure_redirect: TcpListener,
    #[cfg(feature = "secure")]
    tls_acceptor: TlsAcceptor,
}

impl<'a> Config {
    /// Utility that checks if the given mointpoint is already taken. takes in the uri of the to be
    /// mounted mountpoint
    fn check_mountpoint_taken(&self, to_insert: Uri) -> bool {
        if let Some(to_check) = &self.mountpoints {
            for i in to_check.iter() {
                if i.mountpoint == to_insert {
                    return true; // Found a duplicate &str
                }
            }
        };
        false
    }
    /// mounts a [MountPoint] on the [Config] takes in a blank [MountPoint] and the [Route]s to be
    /// mounted, mounts them and inserts the new [MountPoint]. Also sorts by rank.
    pub fn mount(mut self, mountpoint: Uri<'static>, mut routes: Vec<Route<'static>>) -> Self {
        if self.check_mountpoint_taken(mountpoint) {
            eprintln!("\x1b[31mTrying to reassign a mountpoint, mountpoint `{mountpoint}` already taken.\x1b[0m");
            return self;
        }
        routes.sort_by(|a, b| a.rank.cmp(&b.rank));
        let mut mount_message = format!("  >> \x1b[35m{}\x1b[0m\n", mountpoint);
        for (index, route) in routes.iter().enumerate() {
            let indent_sign = match index {
                i if i == routes.len() - 1 => "└─",
                _ => "├─",
            };

            mount_message += &format!(
            "     \x1b[35m{indent_sign}\x1b[0m \x1b[36m(\x1b[0m{}\x1b[36m)\x1b[0m \x1b[32m{}\x1b[0m \x1b[34;4m{}\x1b[24m{}\x1b[0m\n",
                route.name.unwrap_or(""),
                route.method,
                mountpoint,
                route.uri
            )
        }

        println!("{mount_message}");
        let mut temp_mountpoints = None;
        std::mem::swap(&mut self.mountpoints, &mut temp_mountpoints);

        if let Some(mut mountpoints) = temp_mountpoints {
            mountpoints.push(MountPoint { mountpoint, routes });
            self.mountpoints = Some(mountpoints);
        } else {
            self.mountpoints = Some(vec![MountPoint { mountpoint, routes }]);
        }
        self
    }
    /// # Launches/Starts the webserver
    /// Launches a Webserver Configuration
    ///
    /// Is Async
    ///
    /// Is Blocking -> Can be interrupted with ^C
    ///
    /// # Panics
    /// Panics if there are no Mountpoints in the Confiuration to lauch, as well as when the there
    /// is no interrupt signal
    pub async fn launch(self) {
        {
            #[cfg(feature = "secure")] {
                println!(
                    "Server launched from https://{} and http://{}",
                    self.address.local_addr().unwrap(), self.to_secure_redirect.local_addr().unwrap()
                );
            }
            #[cfg(not(feature = "secure"))] {
                println!("Server launched from http://{}", self.address.local_addr().unwrap())
            }
        }
        let mut sigint = signal(SignalKind::interrupt()).unwrap();
        let location_string = format!("Location: https://{}", self.address.local_addr().unwrap());
        loop {
            if !self.selector(&mut sigint, &location_string).await {
                break;
            }
        }
    }
    #[cfg(feature = "secure")]
    async fn selector(&self, sigint: &mut Signal, location_string: &str) -> bool{
        select! {
            _ = sigint.recv() => {
                println!("Shutting down...");
                return false;
            }
            Ok((socket, _)) = self.address.accept() => {
                let mountpoints = self.mountpoints.clone().unwrap();
                let Ok(socket) = self.tls_acceptor.accept(socket).await else { 
                    eprintln!("\x1b[31mClient used http, not https\x1b[0m");
                    return true;
                };
                tokio::spawn(async move { handle_connection(socket, mountpoints).await; });
            }
            Ok((socket, _)) = self.to_secure_redirect.accept() => {
                let redirect_response = Response {
                    headers: vec![location_string.to_string()],
                    cookies: None,
                    status: Some(Status::MovedPermanently),
                    body: Box::new(""),
                };
                tokio::spawn(async move { let _ = redirect_response.write(socket, None).await; });
            }
        }
        true
    }
    #[cfg(not(feature = "secure"))]
    async fn selector(&self, sigint: &mut Signal, _location_string: &str) -> bool {
        select! {
            _ = sigint.recv() => {
                println!("Shutting down...");
                return false;
            }
            Ok((socket, _)) = self.address.accept() => {
                let mountpoints = self.mountpoints.clone().unwrap();
                tokio::spawn(async move { handle_connection(socket, mountpoints).await; });
            }
        }
        true
    }

}

/// # Creates a Webserver Config which can be launched with the launch function
/// Takes the IP and Port as an argument
///
/// Prints out the configuration test
///
/// Is async
///
/// # Example
/// ```
/// async fn example() {
///     let _ = http::build("127.0.0.1:8080", "127.0.0.1:8443");
/// }
/// ```
/// # Panics
/// Panics if the IP is not bindable, or other forms of system errors or it's not a valid
/// IP-Address
#[cfg(feature = "secure")]
pub async fn build(ip: &str, ip_http: &str) -> Config {
    let Ok(listener) = TcpListener::bind(ip).await  else {
        panic!("\x1b[31mCould't bind Listener to address\x1b[0m");
    };
    let ip_vec = ip.splitn(2, ':').collect::<Vec<&str>>();
    if ip_vec.len() != 2 {
        panic!("Invalid IP Address");
    }
    let Ok(listener_http) = TcpListener::bind(ip_http).await  else {
        panic!("\x1b[31mCould't bind Listener to address\x1b[0m");
    };
    let identity = Identity::from_pkcs12(include_bytes!("certificates/identity.pfx"), "1234").unwrap();

    let port = ip_vec[1];
    let ip = ip_vec[0];
    let workers = available_parallelism().unwrap().get();
    println!(
"\x1b[34m⚙ Configuration\x1b[0m
  >> \x1b[34mIp\x1b[0m: {ip}
  >> \x1b[34mPort\x1b[0m: {port}
  >> \x1b[34mWorkers\x1b[0m: {workers}
\x1b[32m Security\x1b[0m
  >> \x1b[32mHttp to Https Redirect: http://{ip_http} -> https://{ip}:{port}\x1b[0m
\x1b[35m🛪 Mountpoints\x1b[0m"
    );
    Config {
        mountpoints: None,
        address: listener,
        to_secure_redirect: listener_http,
        tls_acceptor: native_tls::TlsAcceptor::builder(identity).build().unwrap().into()
    }
}
#[cfg(not(feature = "secure"))]
pub async fn build(ip: &str) -> Config {
    let Ok(listener) = TcpListener::bind(ip).await  else {
        panic!("\x1b[31mCould't bind Listener to address\x1b[0m");
    };
    let ip_vec = ip.splitn(2, ':').collect::<Vec<&str>>();
    if ip_vec.len() != 2 {
        panic!("Invalid IP Address");
    }
    let port = ip_vec[1];
    let ip = ip_vec[0];
    let workers = available_parallelism().unwrap().get();
    println!(
"\x1b[34m⚙ Configuration\x1b[0m
  >> \x1b[34mIp\x1b[0m: {ip}
  >> \x1b[34mPort\x1b[0m: {port}
  >> \x1b[34mWorkers\x1b[0m: {workers}
\x1b[35m🛪 Mountpoints\x1b[0m"
    );
    Config {
        mountpoints: None,
        address: listener,
    }
}
